# IoTHubDeviceClient を用いて IoTEdgeHub を経由せず IoT Hub に接続する場合の使用ライブラリ
from azure.iot.device import IoTHubDeviceClient, Message
import iothub_client
from iothub_client import IoTHubClient, IoTHubClientError, IoTHubTransportProvider, IoTHubClientResult
from iothub_client import IoTHubMessage, IoTHubMessageDispositionResult, IoTHubError, DeviceMethodReturnValue

# IoTHubModeleClient を用いて IoTEdgeHub を経由して IoT Hub に接続する場合の使用ライブラリ
import asyncio
import time
import uuid
from azure.iot.device.aio import IoTHubModuleClient

from logging import getLogger, FileHandler, DEBUG, Formatter
import logging.handlers
from datetime import datetime
import json
import os
import sys

# 本SDKで用いるリターン情報
class return_class:
    def __init__(self):
        self.return_status = "normal"    # "normal"(正常)/"error"(異常)
        self.error_detail = ""           # return_status = "error"の場合の詳細エラー情報

# 本SDKで用いるLogFormatterクラス
class LogFormatter(logging.Formatter):
    converter=datetime.fromtimestamp
    def formatTime(self, record, datefmt=None):
        convert_time = self.converter(record.created)
        if datefmt:
            result_str = convert_time.strftime(datefmt)
        else:
            time_str = convert_time.strftime("%Y-%m-%d %H:%M:%S")
            result_str = "%s.%03d" % (time_str, record.msecs)
        return result_str

# 本SDKで用いる変数オブジェクトクラス
class var_cls:
    def __init__(self):
        # 「RB-SDK-IoT Device 属性情報送信機能 for Python」の設定ファイル"rbdevice.conf"設定値格納オブジェクト
        # "Module":IoTHubModeleClient を用いて IoTEdgeHub を経由して IoT Hub に接続する場合
        # "Device":IoTHubDeviceClient を用いて IoTEdgeHub を経由せず IoT Hub に接続する場合
        self.communicate_method = None
        # communicate_method:"Module"の場合のみ必須。IoT Hub に接続するルートの情報
        self.upload_path = None
        # The device connection string to authenticate the device with your IoT hub.
        # Using the Azure CLI:
        # az iot hub device-identity show-connection-string --hub-name {YourIoTHubName} --device-id MyNodeDevice --output table
        self.CONNECTION_STRING = None
        # ログ情報global変数定義
        self.logger = getLogger(__name__)

class RBDevice:
    """
    「RB-SDK-IoT Device 属性情報送信機能 for Python」
    本SDKはIoT DeviceのユーザープログラムがIoT Deviceの情報をRoboticBaseに送信しようとする場合のRoboticBase標準形式での通信経路を意識しない属性情報送信機能を提供する 
    """

    # ログ設定処理
    def log_setting(self,var_ins):
        """
        ログ情報設定処理
        ログ情報設定ファイルから情報を取得し、ログ設定を実施
        """
        # リターン値初期化
        return_info = return_class()
        return_info.return_status = "normal"
        return_info.error_detail = ""
        
        # ログ設定ファイル名
        setting_file = "logsetting.conf"

        # ログ情報設定項目の初期化（初期値の設定）
        log_propagate = "True"                    # 上位ロガーのハンドラにも展開するか（True）しないか（False）
        log_level = "WARNING"                     # ロガーレベル
        handler_level = "WARNING"                 # ハンドラーレベル
        log_dir = r'./logs'                       # ログファイル格納ディレクトリ名（ない場合は当Bridgeで作成）
        log_filename = "RBDevice.senddevice.log"  # ログファイル名
        log_size = 1000                           # ログ情報設定ファイルからログローテーション用の1ファイルの最大サイズ(kb)
        backup_count = 10                         # ログローテーション用のバックアップファイル作成数

        # ログ情報設定ファイルから情報取得
        json_data = dict()
        try:
            script_dir = os.path.dirname(os.path.abspath(__file__))
            pathsettings = os.path.join(script_dir, "settings", setting_file)

            if os.path.isfile(pathsettings):
                # ログ情報設定ファイルが存在する場合
                with open(pathsettings, 'r') as f:
                    json_data = json.load(f)
            else:
                #return_info.return_status = "error"(続行不能)
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice file {pathsettings} not found'
                return return_info
        except Exception as e:
            #return_info.return_status = "error"(続行不能)
            return_info.return_status = "error"
            return_info.error_detail = rf'RBDevice.senddevice get {pathsettings} failed'
            return return_info

        # ログ情報設定ファイルを取得でき、かつ、該当設定項目がある場合、その値を採用
        if ((return_info.return_status == "normal") and ("log_propagate" in json_data.keys())):    
            log_propagate = json_data["log_propagate"]

        if ((return_info.return_status == "normal") and ("log_level" in json_data.keys())): 
            log_level = json_data["log_level"]

        if ((return_info.return_status == "normal") and ("handler_level" in json_data.keys())): 
            handler_level = json_data["handler_level"]

        if ((return_info.return_status == "normal") and ("log_dir" in json_data.keys())): 
            log_dir = json_data["log_dir"]

        if ((return_info.return_status == "normal") and ("log_filename" in json_data.keys())): 
            log_filename = json_data["log_filename"]

        if ((return_info.return_status == "normal") and ("log_size" in json_data.keys())): 
            log_size = json_data["log_size"]

        if ((return_info.return_status == "normal") and ("backup_count" in json_data.keys())): 
            backup_count = json_data["backup_count"]
        
        try:    
            # ログ設定
            var_ins.logger = getLogger(__name__)
            var_ins.logger.propagate = log_propagate
            var_ins.logger.setLevel(log_level)
            # ログディレクトリがもしなければ作成(階層の途中含め)(存在したらエラーにせず)
            os.makedirs(log_dir, exist_ok=True)
            filename = os.path.join(log_dir, log_filename)
            # ディレクトリ名＋ファイル名のPathオブジェクトをfilename引数に指定
            # ディスク上のログファイルに対するローテーション処理をサポートするRotatingFileHandler クラスを使用
            handler = logging.handlers.RotatingFileHandler(filename, maxBytes=log_size*1024, backup_count=backup_count, mode='a', encoding=None, delay=0)
            # formatter = Formatter('%(asctime)s - %(levelname)s - %(message)s')
            formatter = LogFormatter('%(asctime)s - %(levelname)s - %(message)s')
            handler.setFormatter(formatter)
            handler.setLevel(handler_level)
            var_ins.logger.addHandler(handler)

            return return_info
        except Exception as e:
            tb = sys.exc_info()[2]
            var_ins.logger.error('message:{0}'.format(e.with_traceback(tb)))
            var_ins.logger.error(rf'RBDevice.senddevice logger settings {pathsettings} failed')
            return_info.error_detail = rf'RBDevice.senddevice logger settings {pathsettings} failed'
            
            return return_info

    def rbstandard_check(self,var_ins,status=None,commandstatus=None,battery=None,
      type=None,x=None,y=None,z=None,latitude=None,longitude=None,
      failure=None,logging=None,device_data=None):
        """
        RoboticBase送信標準形式チェック
        本関数はIoT DeviceのユーザープログラムがIoT Deviceの情報をRoboticBaseに送信しようとする場合のRoboticBase標準形式となっていることのチェックを実施する
        チェックでエラーの場合はエラー内容を添えてエラーリターンする
        """
        # リターン値初期化
        return_info = return_class()
        return_info.return_status = "normal"
        return_info.error_detail = ""
        #
        # RoboticBase送信標準形式チェック
        #
        try:
            # status(string)：ステータスを示す文字列、"normal"(正常)/"alarm"(警告)/"error"(異常)/""(未使用) 
            if ((status != None) and
               (not(isinstance(status,str)) or
               ((status != "normal") and
               (status != "alarm") and
               (status != "error") and
               (status != "")))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "status" is invalid'
                return return_info

            # commandStatus(string)：コマンドステータスを示す文字列、"ready"(コマンド実行可能状態)/"busy"(コマンド実行状態)/"alert"(非常事態状態)/
            #                        "connecting"(機器接続試行状態) / "disconnected"(機器接続切断状態) / ""(未使用)
            if ((commandstatus != None) and
               (not(isinstance(commandstatus,str)) or
               ((commandstatus != "ready") and
               (commandstatus != "busy") and
               (commandstatus != "alert") and
               (commandstatus != "connecting") and
               (commandstatus != "disconnected") and
               (commandstatus != "")))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "commandstatus" is invalid'
                return return_info

            # battery(float)：電池残量を`0.0`(0%)～`1.0`(100%)の小数で設定、未使用の場合`0.0`
            if ((battery != None) and
               (not(isinstance(battery,float)) or
               (battery > 1.0) or
               (battery < 0.0))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "battery" is invalid'
                return return_info

            # type(string)：座標タイプ、"rb"(RoboticBase 座標)/"gps"(GPS 座標)/""(未使用)、将来的には"geojson"(将来的)も想定されるが現時点仕様としては対象外とする    
            if ((type != None) and
               (not(isinstance(type,str)) or
               ((type != "rb") and
               (type != "gps") and
               (type != "")))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "type" is invalid'
                return return_info
                
            # x(float)：座標タイプ"RB"の場合の横(x)、原点の位置によってはマイナス値になることもあり得る、未使用の場合`0.0`
            if ((x != None) and
               (not(isinstance(x,float)))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "x" is invalid'
                return return_info

            # y(float)：座標タイプ"RB"の場合の縦(y)、原点の位置によってはマイナス値になることもあり得る、未使用の場合`0.0`
            if ((y != None) and
               (not(isinstance(y,float)))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "y" is invalid'
                return return_info

            # z(float)：座標タイプ"RB"の場合の高さ(z)、原点の位置によってはマイナス値になることもあり得る、未使用の場合`0.0`
            if ((z != None) and
               (not(isinstance(z,float)))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "z" is invalid'
                return return_info

            # latitude(float)：座標タイプ"GPS"の場合の緯度(南緯90°〜0〜北緯90°：－90°〜0〜＋90°)/未使用の場合`0.0`
            if ((latitude != None) and
               (not(isinstance(latitude,float)) or
               (latitude > 90.0) or
               (latitude < -90.0))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "latitude" is invalid'
                return return_info

            # longitude(float)：座標タイプ"GPS"の場合の経度(西経180°〜0〜東経180°：－180°〜0〜＋180°)/未使用の場合`0.0`
            if ((longitude != None) and
               (not(isinstance(longitude,float)) or
               (longitude > 180.0) or
               (longitude < -180.0))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "longitude" is invalid'
                return return_info

            # failure(string)：故障情報を示す文字列、""(正常)/"failure"(異常)"/""(未使用)
            if ((failure != None) and 
               (not(isinstance(failure,str)) or
               ((failure != "") and
               (failure != "failure")))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "failure" is invalid'
                return return_info

            # logging(string)：logging 情報(logging の仕様は、RoboticBase(v1.0)の利用マニュアルに則る) 
            if ((logging != None) and 
               (not(isinstance(logging,str)) or
               ((logging != "") and
               (logging == "")))):
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice argument "logging" is invalid'
                return return_info

            # device_data(object)：IoT Device 固有情報については特段の検査を行わない
            # if (not(isinstance(device_data,(str,int,float,object,dict,list)))):
            #    return_info.error_detail = rf'RBDevice.senddevice argument "device_data" is invalid'
            #    return return_info

            # ここまでのチェックでエラーが無ければ、全ての引数が正常のため、正常リターン

            return return_info

        except Exception as e:
            tb = sys.exc_info()[2]
            var_ins.logger.error('message:{0}'.format(e.with_traceback(tb)))
            var_ins.logger.error(rf'RBDevice.senddevice rbstandard_check Exception occured')
            return_info.return_status = "error"
            return_info.error_detail = rf'RBDevice.senddevice rbstandard_check Exception occured'
            
            return return_info

    def iothub_client_init(self,var_ins):
        """
        IoT Hub接続設定値取得および接続準備
        本SDK設定ファイルを取得し、取得内容によって、IoTHubModeleClient を用いて IoTEdgeHub を経由して IoT Hub に接続するか、
        IoTHubDeviceClient を用いて IoTEdgeHub を経由せず IoT Hub に接続するための設定値を取得し、接続準備を行う。
        """
        # リターン値初期化
        return_info = return_class()
        return_info.return_status = "normal"
        return_info.error_detail = ""
        # RoboticBase送信設定ファイル名
        setting_file = "rbdevice.conf"
        # 設定ファイルから情報取得
        json_data = dict()
        try:
            script_dir = os.path.dirname(os.path.abspath(__file__))
            if os.path.isfile(os.path.join(script_dir, "settings", setting_file)):
                # MQTT送信情報設定ファイルが存在する場合
                with open(os.path.join(script_dir, "settings", setting_file), 'r') as f:
                    json_data = json.load(f)
                    if ("communicate_method" in json_data.keys()):
                        var_ins.communicate_method = json_data["communicate_method"]
                    else:
                        var_ins.logger.error(rf'RBDevice.senddevice ./settings/{setting_file} get failed')
                        return_info.return_status = "error"
                        return_info.error_detail = rf'RBDevice.senddevice ./settings/{setting_file} get failed'
                        return return_info
                    
                    # communicate_method："Module":IoTHubModeleClient を用いて接続する場合"upload_path"の指定がなければエラーリターン
                    if (var_ins.communicate_method == "Module"):
                        if ("upload_path" in json_data.keys()):
                            var_ins.upload_path = json_data["upload_path"]
                        else:
                            var_ins.logger.error(rf'RBDevice.senddevice ./settings/{setting_file} invalid')
                            return_info.return_status = "error"
                            return_info.error_detail = rf'RBDevice.senddevice ./settings/{setting_file} invalid'
                            return return_info

                    # communicate_method："Device":IoTHubDeviceClient を用いて接続する場合"rb-device-connection_string"の指定がなければエラーリターン
                    if (var_ins.communicate_method == "Device"):
                        if ("rb_device_connection_string" in json_data.keys()):
                            var_ins.CONNECTION_STRING = json_data["rb_device_connection_string"]
                        else:
                            var_ins.logger.error(rf'RBDevice.senddevice ./settings/{setting_file} invalid')
                            return_info.return_status = "error"
                            return_info.error_detail = rf'RBDevice.senddevice ./settings/{setting_file} invalid'
                            return return_info

                    # communicate_methodが"Module"でも"Device"でもない場合、エラーリターン
                    if (not(var_ins.communicate_method == "Module") and
                       not(var_ins.communicate_method == "Device")):
                        var_ins.logger.error(rf'RBDevice.senddevice ./settings/{setting_file} invalid')
                        return_info.return_status = "error"
                        return_info.error_detail = rf'RBDevice.senddevice ./settings/{setting_file} invalid'
                        return return_info                        

            var_ins.logger.info(rf'RBDevice.senddevice ./settings/{setting_file} get succeeded')

            # communicate_method："Device":IoTHubDeviceClient を用いて接続する場合
            if ((return_info.return_status == "normal") and
               (var_ins.communicate_method == "Device")):

                # IoTHubDeviceClientを生成して接続
                client = IoTHubDeviceClient.create_from_connection_string(var_ins.CONNECTION_STRING)
                
                var_ins.logger.info(rf'RBDevice.senddevice Create an IoT Hub Device Client succeeded')
                return return_info,client

            # communicate_method："Module":IoTHubModeleClient を用いて接続する場合
            if ((return_info.return_status == "normal") and
               (var_ins.communicate_method == "Module")):
      
                # Inputs/Ouputs are only supported in the context of Azure IoT Edge and module client
                # The module client object acts as an Azure IoT Edge module and interacts with an Azure IoT Edge hub
                # IoTHubModeleClientを生成して接続
                module_client = IoTHubModuleClient.create_from_edge_environment()

                # Connect the client
                #await module_client.connect()
                module_client.connect()
                
                var_ins.logger.info(rf'RBDevice.senddevice Create an IoT Hub Module Client succeeded')
                return return_info,module_client

        except Exception as e:
            tb = sys.exc_info()[2]
            var_ins.logger.error('message:{0}'.format(e.with_traceback(tb)))
            var_ins.logger.error(rf'RBDevice.senddevice Create an IoT Hub Module/Device Client failed')
            return_info.return_status = "error"
            return_info.error_detail = rf'RBDevice.senddevice Create an IoT Hub Module/Device Client failed'
           
            return return_info

    #
    # IoT Device情報RoboticBase送信
    #
    def senddevice(self,status=None,commandstatus=None,battery=None,
      type=None,x=None,y=None,z=None,latitude=None,longitude=None,
      failure=None,logging=None,device_data=None):
        """
        IoT Device情報RoboticBase送信
        本関数はIoT DeviceのユーザープログラムがIoT Deviceの情報をRoboticBaseに送信しようとする場合のRoboticBase標準形式での通信経路を意識しない属性情報送信機能を提供する
        """
        # リターン値初期化
        return_info = return_class()
        return_info.return_status = "normal"
        return_info.error_detail = ""

        # 本Azure Functions関数で用いる変数オブジェクトクラスのインスタンス化
        var_ins = var_cls()
        #
        # ログ設定処理
        #
        try:
            return_info = self.log_setting(var_ins)

            if (return_info.return_status == "normal"):
                  
                # 引数ロギング
                var_ins.logger.info(rf'RBDevice.senddevice status=[{status}]')
                var_ins.logger.info(rf'RBDevice.senddevice commandstatus=[{commandstatus}]')
                var_ins.logger.info(rf'RBDevice.senddevice battery=[{battery}]')
                var_ins.logger.info(rf'RBDevice.senddevice type=[{type}]')
                var_ins.logger.info(rf'RBDevice.senddevice x=[{x}]')
                var_ins.logger.info(rf'RBDevice.senddevice y=[{y}]')
                var_ins.logger.info(rf'RBDevice.senddevice z=[{z}]')
                var_ins.logger.info(rf'RBDevice.senddevice latitude=[{latitude}]')
                var_ins.logger.info(rf'RBDevice.senddevice longitude=[{longitude}]')
                var_ins.logger.info(rf'RBDevice.senddevice failure=[{failure}]')
                var_ins.logger.info(rf'RBDevice.senddevice logging=[{logging}]')
                var_ins.logger.info(rf'RBDevice.senddevice device_data=[{device_data}]')
                #
            else:
                var_ins.logger.error(rf'RBDevice.senddevice log_setting failed')
                return return_info    

        except Exception as e:
            tb = sys.exc_info()[2]
            var_ins.logger.error('message:{0}'.format(e.with_traceback(tb)))
            var_ins.logger.error(rf'RBDevice.senddevice log_setting failed')
            return return_info

        if (return_info.return_status == "normal"):
            try:
                #
                # RoboticBase送信標準形式チェック処理
                #
                return_info = self.rbstandard_check(var_ins,status=status,commandstatus=commandstatus,battery=battery,
                  type=type,x=x,y=y,z=z,latitude=latitude,longitude=longitude,
                  failure=failure, logging=logging, device_data=device_data)
        
            except Exception as e:
                tb = sys.exc_info()[2]
                var_ins.logger.error('message:{0}'.format(e.with_traceback(tb)))
                var_ins.logger.error(rf'RBDevice.senddevice Exception occured')
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice Exception occured'
                return return_info

        if (return_info.return_status == "normal"):
            try:
                # IoT Hub Client初期化処理
                return_info,client = self.iothub_client_init(var_ins)

                # RoboticBase送信標準形式データ編集(device_dataの有無で送信データを編集)
                # 2020/02/20 機能横断チームの「RB-SDK-IoT Device 属性情報送信機能 for Python」仕様レビュー指摘に従い
                # 呼び出し元から設定されなかったRoboticBase送信項目はRoboticBaseに送信しないようプログラム変更

                message_str = None

                if (status != None):
                    message_str = r'{"status": '+rf'"{status}"'

                if ((message_str != None) and
                   (commandstatus != None)):
                    message_str = rf'{message_str}'+r',"commandstatus": '+rf'"{commandstatus}"'
                else:
                    message_str = r'{"commandstatus": '+rf'"{commandstatus}"'

                if ((message_str != None) and
                   (battery != None)):
                    message_str = rf'{message_str}'+r',"battery": '+rf'{battery}'
                else:
                    message_str = r'{"battery": '+rf'{battery}'

                if ((message_str != None) and
                   (type != None)):
                    message_str = rf'{message_str}'+r',"type": '+rf'"{type}"'
                else:
                    message_str = r'{"type": '+rf'"{type}"'

                if ((message_str != None) and
                   (x != None)):
                    message_str = rf'{message_str}'+r',"x": '+rf'{x}'
                else:
                    message_str = r'{"x": '+rf'{x}'

                if ((message_str != None) and
                   (y != None)):
                    message_str = rf'{message_str}'+r',"y": '+rf'{y}'
                else:
                    message_str = r'{"y": '+rf'{y}'

                if ((message_str != None) and
                   (z != None)):
                    message_str = rf'{message_str}'+r',"z": '+rf'{z}'
                else:
                    message_str = r'{"z": '+rf'{z}'

                if ((message_str != None) and
                   (latitude != None)):
                    message_str = rf'{message_str}'+r',"latitude": '+rf'{latitude}'
                else:
                    message_str = r'{"latitude": '+rf'{latitude}'

                if ((message_str != None) and
                   (longitude != None)):
                    message_str = rf'{message_str}'+r',"longitude": '+rf'{longitude}'
                else:
                    message_str = r'{"longitude": '+rf'{longitude}'

                if ((message_str != None) and
                   (failure != None)):
                    message_str = rf'{message_str}'+r',"failure": '+rf'"{failure}"'
                else:
                    message_str = r'{"failure": '+rf'"{failure}"'

                if ((message_str != None) and
                   (logging != None)):
                    message_str = rf'{message_str}'+r',"logging": '+rf'{logging}'
                else:
                    message_str = r'{"logging": '+rf'{logging}'

                if ((message_str != None) and
                   (device_data != None)):
                    message_str = rf'{message_str}'+r',"device_data": '+rf'{device_data}'
                else:
                    message_str = r'{"device_data": '+rf'{device_data}'

                if (message_str != None):
                    message_str = rf'{message_str}'+r'}'

                message = message_str

                var_ins.logger.info(rf'RBDevice.senddevice after edited message:{message}')

                # communicate_method："Device":IoTHubDeviceClient を用いて接続する場合
                if ((return_info.return_status == "normal") and
                   (var_ins.communicate_method == "Device")):
                
                    var_ins.logger.info(rf'RBDevice.senddevice before client.message:{message}')

                    client.send_message(message)

                    var_ins.logger.info(rf'RBDevice.senddevice client.message:{message} succeeded')

                    client.disconnect()

                    var_ins.logger.info(rf'RBDevice.senddevice client.disconnect() succeeded')

                # communicate_method："Module":IoTHubModeleClient を用いて接続する場合
                if ((return_info.return_status == "normal") and
                   (var_ins.communicate_method == "Module")):
                    
                    var_ins.logger.info(rf'RBDevice.senddevice before client.send_message_to_output:{message}')

                    #await client.send_message_to_output(message, var_ins.upload_path)
                    client.send_message_to_output(message, var_ins.upload_path)

                    var_ins.logger.info(rf'RBDevice.senddevice client.send_message_to_output:{message} succeeded')

                    client.disconnect()

                    var_ins.logger.info(rf'RBDevice.senddevice client.disconnect() succeeded')

                return return_info

            except Exception as e:
                tb = sys.exc_info()[2]
                var_ins.logger.error('message:{0}'.format(e.with_traceback(tb)))
                var_ins.logger.error(rf'RBDevice.senddevice client.message:{message} failed')
                return_info.return_status = "error"
                return_info.error_detail = rf'RBDevice.senddevice client.message:{message} failed'

                return return_info
        else:
            var_ins.logger.error(error_detail)

            return return_info